# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Module that has models for version control systems, for inspektor scripts.
"""
import logging
import os

from inspektor import exceptions
from inspektor import utils
from inspektor.utils import process


class VCS(object):

    """
    Abstraction layer to the version control system.
    """

    def __init__(self):
        """
        Guess the version control system and instantiate it.
        """
        self.log = logging.getLogger("inspektor.vcs")
        self.backend = self.get_backend()
        self.cwd = os.getcwd()

    def get_backend(self):
        if os.path.isdir(".svn"):
            return SubVersionBackend()
        elif os.path.exists(".git"):
            return GitBackend()
        else:
            self.log.error("Could not figure version control system. You "
                           "must be at the top of the project's directory")
            return 1

    def get_unknown_files(self):
        """
        Return a list of files unknown to the VCS.
        """
        return self.backend.get_unknown_files()

    def get_modified_files(self):
        """
        Return a list of files that were modified, according to the VCS.
        """
        return self.backend.get_modified_files()

    def is_file_tracked(self, fl):
        """
        Return whether a file is tracked by the VCS.
        """
        return self.backend.is_file_tracked(fl)

    def add_untracked_file(self, fl):
        """
        Add an untracked file to version control.
        """
        return self.backend.add_untracked_file(fl)

    def revert_file(self, fl):
        """
        Restore file according to the latest state on the reference repo.
        """
        return self.backend.revert_file(fl)

    def set_file_executable(self, fl):
        """
        Set executable permissions for file inside version control.
        """
        return self.backend.set_file_executable(fl)

    def unset_file_executable(self, fl):
        """
        Set executable permissions for file inside version control.
        """
        return self.backend.unset_file_executable(fl)

    def apply_patch(self, patch):
        """
        Applies a patch using the most appropriate method to the particular VCS.
        """
        return self.backend.apply_patch(patch)

    def update(self):
        """
        Updates the tree according to the latest state of the public tree
        """
        return self.backend.update()

    def get_modified_files_patch(self, untracked_files_before, patch):
        """
        Get modified files after a patch application.

        :param untracked_files_before: List of files untracked previous to
                patch application.
        """
        return self.backend.get_modified_files_patch(untracked_files_before,
                                                     patch)


class SubVersionBackend(object):

    """
    Implementation of a subversion backend for use with the VCS layer.
    """

    def __init__(self):
        self.log = logging.getLogger('inspektor.vcs')
        self.name = 'subversion'
        self.ignored_extension_list = ['.orig', '.bak']

    def get_unknown_files(self):
        result = process.run("svn status --ignore-externals", verbose=False)
        unknown_files = []
        for line in result.stdout.split("\n"):
            status_flag = line.split()[0]
            if line and status_flag == "?":
                for extension in self.ignored_extension_list:
                    if not line.endswith(extension):
                        unknown_files.append(line[1:].strip())
        return unknown_files

    def get_modified_files(self):
        result = process.run("svn status --ignore-externals", verbose=False)
        modified_files = []
        for line in result.stdout.split("\n"):
            status_flag = line.split()[0]
            if line and status_flag == "M" or status_flag == "A":
                modified_files.append(line[1:].strip())
        return modified_files

    def is_file_tracked(self, fl):
        stdout = None
        try:
            result = process.run("svn status --ignore-externals %s" % fl,
                                 verbose=False)
            stdout = result.stdout
        except exceptions.CmdError:
            return False

        if stdout is not None:
            if stdout:
                status_flag = stdout.split()[0]
                if status_flag == "?":
                    return False
                else:
                    return True
            else:
                return True
        else:
            return False

    def add_untracked_file(self, fl):
        """
        Add an untracked file under revision control.

        :param file: Path to untracked file.
        """
        try:
            process.run('svn add %s' % fl, verbose=False)
        except exceptions.CmdError, e:
            self.log.error("Problem adding file %s to svn: %s", fl, e)
            return 1

    def revert_file(self, fl):
        """
        Revert file against last revision.

        :param file: Path to file to be reverted.
        """
        try:
            process.run('svn revert %s' % fl, verbose=False)
        except exceptions.CmdError, e:
            self.log.error("Problem reverting file %s: %s", fl, e)
            return 1

    def set_file_executable(self, fl):
        """
        Set executable permissions for file inside version control.
        """
        process.run("svn propset svn:executable ON %s" % fl,
                    ignore_status=True)

    def unset_file_executable(self, fl):
        """
        Unset executable permissions for file inside version control.
        """
        process.run("svn propdel svn:executable %s" % fl,
                    ignore_status=True)

    def apply_patch(self, patch):
        """
        Apply a patch to the code base. Patches are expected to be made using
        level -p1, and taken according to the code base top level.

        :param patch: Path to the patch file.
        """
        try:
            process.run("patch -p1 < %s" % patch, verbose=False)
        except:
            self.log.error("Patch applied incorrectly. Possible causes: ")
            self.log.error("1 - Patch might not be -p1")
            self.log.error("2 - You are not at the top of the tree")
            self.log.error("3 - Patch was made using an older tree")
            self.log.error("4 - Mailer might have messed the patch")
            return 1

    def update(self):
        try:
            process.run("svn update")
        except exceptions.CmdError, e:
            self.log.error("SVN tree update failed: %s" % e)

    def get_modified_files_patch(self, untracked_files_before, patch):
        """
        Get modified files after a patch application.

        :param untracked_files_before: List of files untracked previous to
                patch application.
        :param patch: Path to the patch file. Unused in the svn implementation.
        """
        untracked_files = self.vcs.get_unknown_files()
        modified_files = self.vcs.get_modified_files()
        add_to_vcs = []
        for untracked_file in untracked_files:
            if untracked_file not in untracked_files_before:
                add_to_vcs.append(untracked_file)

        if add_to_vcs:
            self.log.info("The files: ")
            for untracked_file in add_to_vcs:
                self.log.info(untracked_file)
            self.log.info("Might need to be added to svn")
            answer = utils.ask("Would you like to add them to svn ?")
            if answer == "y":
                for untracked_file in add_to_vcs:
                    self.add_untracked_file(untracked_file)
                    modified_files.append(untracked_file)
            elif answer == "n":
                pass

        return modified_files


class GitBackend(object):

    """
    Implementation of a git backend for use with the VCS abstraction layer.
    """

    def __init__(self):
        self.ignored_extension_list = ['.orig', '.bak']
        self.name = 'git'
        self.log = logging.getLogger("inspektor.vcs")

    def get_unknown_files(self):
        result = process.run("git status --porcelain", verbose=False)
        unknown_files = []
        for line in result.stdout.split("\n"):
            if line:
                status_flag = line.split()[0]
                if status_flag == "??":
                    for extension in self.ignored_extension_list:
                        if not line.endswith(extension):
                            element = line[2:].strip()
                            if element not in unknown_files:
                                unknown_files.append(element)
        return unknown_files

    def get_modified_files(self):
        result = process.run("git status --porcelain", verbose=False)
        modified_files = []
        for line in result.stdout.split("\n"):
            if line:
                status_flag = line.split()[0]
                if status_flag in ["M", "A"]:
                    modified_files.append(line.split()[-1])
        return modified_files

    def is_file_tracked(self, fl):
        try:
            process.run("git ls-files %s --error-unmatch" % fl, verbose=False)
            return True
        except exceptions.CmdError:
            return False

    def add_untracked_file(self, fl):
        """
        Add an untracked file under revision control.

        :param file: Path to untracked file.
        """
        try:
            process.run('git add %s' % fl, verbose=False)
        except exceptions.CmdError, e:
            self.log.error("Problem adding file %s to git: %s", fl, e)
            return 1

    def revert_file(self, fl):
        """
        Revert file against last revision.

        :param file: Path to file to be reverted.
        """
        try:
            process.run('git checkout %s' % fl, verbose=False)
        except exceptions.CmdError, e:
            self.log.error("Problem reverting file %s: %s", fl, e)
            return 1

    def set_file_executable(self, fl):
        """
        Set executable permissions for file inside version control.
        """
        process.run("chmod +x %s" % fl, ignore_status=True)

    def unset_file_executable(self, fl):
        """
        Unset executable permissions for file inside version control.
        """
        process.run("chmod -x %s" % fl, ignore_status=True)

    def apply_patch(self, patch):
        """
        Apply a patch to the code base using git am.

        A new branch will be created with the patch name.

        :param patch: Path to the patch file.
        """
        def branch_suffix(name):
            suffix = 1
            while not process.run("git show-ref --verify --quiet "
                                  "refs/heads/%s_%s" % (branch, suffix),
                                  verbose=False,
                                  ignore_status=True).exit_status:
                suffix += 1
            return "%s_%s" % (name, suffix)
        process.run("git checkout master", verbose=False)
        branch = os.path.basename(patch).rstrip(".patch")
        try:
            process.run("git checkout -b %s" % branch,
                        verbose=False)
        except exceptions.CmdError:
            self.log.error("branch %s already exists!"
                           % branch)
            answer = utils.ask("What would you like to do?",
                               options="Abort/Delete/Rename/OldBase/NewBase")
            if not answer:
                answer = "A"
            answer = answer[0].upper()
            if answer == "A":
                self.log.info("Aborting check")
                return 1
            elif answer == "D":
                self.log.info("Deleting branch %s", branch)
                process.run("git branch -D %s" % branch, verbose=False)
                process.run("git checkout -b %s" % branch, verbose=False)
            elif answer == "R":     # Rename the old branch
                old_branch = branch_suffix(branch)
                self.log.info("Moving branch %s to %s", branch, old_branch)
                process.run("git branch -M %s %s"
                            % (branch, old_branch), verbose=False)
                process.run("git checkout -b %s" % branch, verbose=False)
            elif answer == "O":     # Rename the old branch and use old master
                old_branch = branch_suffix(branch)
                self.log.info("Moving branch %s to %s", branch, old_branch)
                process.run("git branch -M %s %s"
                            % (branch, old_branch), verbose=False)
                process.run("git checkout -b %s" % branch, verbose=False)
                base = process.run("git merge-base %s %s"
                                   % (branch, old_branch),
                                   verbose=False).stdout.strip()
                self.log.info("Last common base is %s", base)
                process.run("git reset --hard %s" % base, verbose=False)
            elif answer == "N":     # Rename and rebase the old branch
                old_branch = branch_suffix(branch)
                self.log.info("Moving branch %s to %s and rebasing it to "
                              "the current master", branch, old_branch)
                process.run("git branch -M %s %s"
                            % (branch, old_branch), verbose=False)
                process.run("git checkout %s" % old_branch, verbose=False)
                ret = process.run("git rebase master", verbose=False,
                                  ignore_status=True)
                if ret.exit_status:
                    self.log.error("Fail to automatically rebase old branch "
                                   "%s to the current master. Continuing "
                                   "without automatic rebase (as if you'd "
                                   "chosen 'Rename')", old_branch)
                    process.run("git rebase --abort", verbose=False,
                                ignore_status=True)
                process.run("git checkout master", verbose=False)
                process.run("git checkout -b %s" % branch, verbose=False)
        try:
            process.run("git am -3 %s" % patch, verbose=False)
        except exceptions.CmdError, e:
            self.log.error("Failed to apply patch to the git repo: %s" % e)
            return 1

    def update(self):
        try:
            process.run("git pull", verbose=False)
        except exceptions.CmdError, e:
            self.log.error("git tree update failed: %s" % e)

    def get_modified_files_patch(self, untracked_files_before, patch):
        """
        Get modified files after a patch application.

        :param untracked_files_before: List of files untracked before patch
                application. Unused in the git implementation because git is
                awesome and we really don't need to keep track of them. This
                is here just for symmetry with subversion.
        :param patch: Path to the patch file.
        """
        modified_files = []
        patch_contents = open(patch, 'r')
        for line in patch_contents.readlines():
            if line.startswith("diff --git"):
                m_file = line.split()[-1][2:]
                if m_file not in modified_files:
                    modified_files.append(m_file)
        patch_contents.close()
        return modified_files



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

import logging
import os
import tempfile

from inspektor import lint
from inspektor import reindent
from inspektor import style
from inspektor import inspector
from inspektor import vcs
from inspektor import utils

TMP_FILE_DIR = tempfile.gettempdir()
LOG_FILE_PATH = os.path.join(TMP_FILE_DIR, 'check-patch.log')
# Hostname of patchwork server to use
PWHOST = "patchwork.virt.bos.redhat.com"

log = logging.getLogger("inspektor.check")

# Rely on built-in recursion limit to limit number of directories searched


def license_project_name(path):
    '''
    Locate the nearest LICENSE file, take first word as the project name
    '''
    if path == '/' or path == '.':
        raise RuntimeError('Ran out of directories searching for LICENSE file')
    try:
        license_file = file(os.path.join(path, 'LICENSE'), 'r')
        first_word = license_file.readline().strip().split()[0].lower()
        return first_word, path
    except IOError:
        # Recurse search parent of path's directory
        return license_project_name(os.path.dirname(path))


class FileChecker(object):

    """
    Picks up a given file and performs various checks, looking after problems
    and eventually applying solutions when all possible.
    """

    def __init__(self):
        """
        Class constructor, sets the file checkers.

        :param path: Path to the file that will be checked.
        :param vcs: Version control system being used.
        :param confirm: Whether to answer yes to all questions asked without
                prompting the user.
        """
        self.linter = lint.Linter(verbose=False)
        self.indenter = reindent.Reindenter(verbose=False)
        self.style_checker = style.StyleChecker()
        self.vcs = vcs.VCS()

    def _check_indent(self, path):
        """
        Verifies the file with the reindent module.

        This tool performs the following checks on python files:

          * Trailing whitespaces
          * Tabs
          * End of line
          * Incorrect indentation

        And will automatically correct problems.
        """
        return self.indenter.check_file(path)

    def _check_syntax(self, path):
        """
        Verifies the file with pylint.
        """
        return self.linter.check_file(path)

    def _check_style(self, path):
        """
        Verifies the file compliance to PEP8.
        """
        return self.style_checker.check_file(path)

    def _check_permissions(self, path):
        """
        Verifies the execution permissions and fixes them if possible:

          * Files with no shebang and execution permissions will be fixed.
          * Files with shebang and no execution permissions will be fixed.
        """
        result = True
        if os.path.isdir(path):
            return result
        path_inspector = inspector.PathInspector(path)
        path_is_script = path_inspector.is_script()
        path_is_exec = path_inspector.has_exec_permission()
        if path_is_script:
            if not path_is_exec:
                result = False
                self.vcs.set_file_executable(path)
        else:
            if path_is_exec:
                result = False
                self.vcs.unset_file_executable(path)
        return result

    def check_file(self, path):
        """
        Executes all required checks, if problems are found, the possible
        corrective actions are listed.
        """
        return (self._check_syntax(path) and
                self._check_indent(path) and
                self._check_style(path) and
                self._check_permissions(path))


class PatchChecker(FileChecker):

    def __init__(self, patch=None, patchwork_id=None, github_id=None):
        FileChecker.__init__(self)
        self.base_dir = TMP_FILE_DIR

        if patch:
            self.patch = os.path.abspath(patch)

        if patchwork_id:
            self.patch = self._fetch_from_patchwork(patchwork_id)

        if github_id:
            self.patch = self._fetch_from_github(github_id)

    def validate(self):
        if not os.path.isfile(self.patch):
            logging.error("Invalid patch file %s provided. Aborting.",
                          self.patch)
            return 1

        changed_files_before = self.vcs.get_modified_files()
        if changed_files_before:
            log.error("Repository has changed files prior to patch "
                      "application")
            answer = utils.ask("Would you like to revert them?")
            if answer == "n":
                log.error("Not safe to proceed without reverting files.")
                return 1
            else:
                for changed_file in changed_files_before:
                    self.vcs.revert_file(changed_file)

        self.untracked_files = self.vcs.get_unknown_files()

    def _get_patchwork_url(self, pw_id):
        return "http://%s/patch/%s/mbox/" % (self.pwhost, pw_id)

    def _download_from_patchwork(self, pw_id):
        patch_url = self._get_patchwork_url(pw_id)
        patch_dest = os.path.join(self.base_dir, 'patchwork-%s.patch' %
                                  pw_id)
        return utils.download.get_file(patch_url, patch_dest)

    def _fetch_from_patchwork(self, pw_id):
        """
        Gets a patch file from patchwork and puts it under the cwd so it can
        be applied.

        :param pw_id: Patchwork patch id. It can be a string with
                comma separated patchwork ids.
        """
        collection = os.path.join(self.base_dir, 'patchwork-%s.patch' %
                                  utils.random_string(4))
        collection_rw = open(collection, 'w')

        for p_id in pw_id.split(","):
            patch = self._download_from_patchwork(p_id)
            # Patchwork sometimes puts garbage on the path, such as long
            # sequences of underscores (_______). Get rid of those.
            patch_ro = open(patch, 'r')
            patch_contents = patch_ro.readlines()
            patch_ro.close()
            for line in patch_contents:
                if not line.startswith("___"):
                    collection_rw.write(line)
        collection_rw.close()

        return collection

    def _get_github_project_name(self):
        project_name, _ = license_project_name(self.vcs.cwd)
        return project_name

    def _get_github_url(self, gh_id):
        """
        Gets the correct github URL for the given project.
        """
        return ("https://github.com/autotest/%s/pull/%s.patch" %
                (self._get_github_project_name(), gh_id))

    def _fetch_from_github(self, gh_id):
        """
        Gets a patch file from patchwork and puts it under the base dir.

        :param gh_id: Patchwork patch id.
        """
        patch_url = self._get_github_url(gh_id)
        patch_dest = os.path.join(self.base_dir, 'github-%s.patch' % gh_id)
        return utils.download.get_file(patch_url, patch_dest)

    def _check_files_modified_patch(self):
        files_failed_check = []

        modified_files = self.vcs.get_modified_files_patch(self.untracked_files,
                                                           self.patch)
        for modified_file in modified_files:
            # Additional safety check, new commits might introduce
            # new directories
            if os.path.isfile(modified_file):
                if not self.check_file(modified_file):
                    files_failed_check.append(modified_file)

        if files_failed_check:
            return False
        else:
            return True

    def check(self):
        self.vcs.apply_patch(self.patch)
        return self._check_files_modified_patch()


def set_arguments(parser):
    pgh = parser.add_parser('github',
                            help='check GitHub Pull Requests')
    pgh.add_argument('gh_id', type=int,
                     help='GitHub Pull Request ID')
    pgh.set_defaults(func=check_patch_github)


def check_patch_github(args):
    gh_id = args.gh_id
    checker = PatchChecker(github_id=gh_id)
    checker.validate()
    if checker.check():
        log.info("Github ID #%s check PASS", gh_id)
        return 0
    else:
        log.info("Github ID #%s check FAIL", gh_id)
        return 1



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

import logging
import os
from inspector import PathInspector

log = logging.getLogger("inspektor.license")

LICENSE_SNIPPET_GPLV2 = """# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
"""

LICENSE_SNIPPET_GPLV2_STRICT = """# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; specifically version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
"""

license_mapping = {'gplv2_later': LICENSE_SNIPPET_GPLV2,
                   'gplv2_strict': LICENSE_SNIPPET_GPLV2_STRICT}
default_license = 'gplv2_later'


class LicenseChecker(object):

    def __init__(self, license_type=default_license, cpyright="", author=""):
        self.failed_paths = []

        self.license_contents = license_mapping[license_type]
        self.base_license_contents = self.license_contents

        if cpyright:
            self.license_contents += "#\n"
            self.license_contents += "# " + cpyright + "\n"

        if author:
            self.license_contents += "# " + author + "\n"

    def check_dir(self, path):
        def visit(arg, dirname, filenames):
            for filename in filenames:
                self.check_file(os.path.join(dirname, filename))

        os.path.walk(path, visit, None)
        return not self.failed_paths

    def check_file(self, path):
        inspector = PathInspector(path)
        if inspector.is_toignore():
            return True
        # Don't put license info in empty __init__.py files.
        if not inspector.is_python() or inspector.is_empty():
            return True

        first_line = None
        if inspector.is_script("python"):
            first_line = inspector.get_first_line()

        new_content = None
        with open(path, 'r') as inspected_file:
            content = inspected_file.readlines()
            if first_line is not None:
                content = content[1:]
            content = "".join(content)
            if self.base_license_contents not in content:
                new_content = ""
                if first_line is not None:
                    new_content += first_line
                    new_content += '\n'
                new_content += self.license_contents + '\n' + content

        if new_content is not None:
            with open(path, 'w') as inspected_file:
                inspected_file.write(new_content)
                self.failed_paths.append(path)
                return False

        return True

    def check(self, path):
        if os.path.isfile(path):
            return self.check_file(path)
        elif os.path.isdir(path):
            return self.check_dir(path)


def set_arguments(parser):
    plicense = parser.add_parser('license',
                                 help='check for presence of license files')
    plicense.add_argument('path', type=str,
                          help='Path to check (empty for full tree check)',
                          nargs='?',
                          default="")
    plicense.add_argument('--license', type=str,
                          help=('License type. Supported license types: %s. '
                                'Default: %s' %
                                (license_mapping.keys(), default_license)),
                          default="gplv2_later")
    plicense.add_argument('--copyright', type=str,
                          help='Copyright string. Ex: "Copyright (c) 2013-2014 FooCorp"',
                          default="")
    plicense.add_argument('--author', type=str,
                          help='Author string. Ex: "Author: Brandon Lindon <brandon.lindon@foocorp.com>"',
                          default="")
    plicense.set_defaults(func=run_license)


def run_license(args):
    path = args.path
    license_type = args.license
    cpyright = args.copyright
    author = args.author

    if not path:
        path = os.getcwd()

    checker = LicenseChecker(license_type=license_type,
                             cpyright=cpyright, author=author)

    if checker.check(path):
        log.info("License check PASS")
        return 0
    else:
        log.error("License check FAIL")
        return 1



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

import logging
import random
import string

import download
import process

log = logging.getLogger('inspektor.utils')


def ask(question, auto=False, options="y/n"):
    """
    Raw input with a prompt.

    :param question: Question to be asked
    :param auto: Whether to return "y" instead of asking the question
    """
    if auto:
        log.info("%s (%s) y" % (question, options))
        return "y"
    return raw_input("%s (%s) " % (question, options))


def random_string(length, ignore_str=string.punctuation,
                  convert_str=""):
    """
    Return a random string using alphanumeric characters.

    :param length: Length of the string that will be generated.
    :param ignore_str: Characters that will not include in generated string.
    :param convert_str: Characters that need to be escaped (prepend "\\").

    :return: The generated random string.
    """
    r = random.SystemRandom()
    result = ""
    chars = string.letters + string.digits + string.punctuation
    if not ignore_str:
        ignore_str = ""
    for i in ignore_str:
        chars = chars.replace(i, "")

    while length > 0:
        tmp = r.choice(chars)
        if convert_str and (tmp in convert_str):
            tmp = "\\%s" % tmp
        result += tmp
        length -= 1
    return str



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

import logging
import os
import sys

from pylint.lint import Run

from inspector import PathInspector

log = logging.getLogger("inspektor.lint")


class Linter(object):

    def __init__(self, verbose=True):
        self.verbose = verbose
        self.ignored_errors = 'E1002,E1101,E1103,E1120,F0401,I0011'
        # Be able to analyze all imports inside the project
        sys.path.insert(0, os.getcwd())
        self.failed_paths = []

    def set_verbose(self):
        self.verbose = True

    def get_opts(self):
        """
        If VERBOSE is set, show all complaints. If not, only errors.
        """
        if not self.verbose:
            return ['--disable=W,R,C,%s' % self.ignored_errors,
                    '--reports=no', '--rcfile=/dev/null',
                    '--good-names=i,j,k,Run,_,vm',
                    '--msg-template="{msg_id}:{line:3d},{column}: {obj}: {msg}"']
        else:
            return []

    def check_dir(self, path):
        """
        Recursively go on a directory checking files with pylint.

        :param path: Path to a directory.
        """
        def visit(arg, dirname, filenames):
            for filename in filenames:
                self.check_file(os.path.join(dirname, filename))

        os.path.walk(path, visit, None)
        return not self.failed_paths

    def check_file(self, path):
        """
        Check one regular file with pylint for py syntax errors.

        :param path: Path to a regular file.
        :return: False, if pylint found syntax problems, True, if pylint didn't
                 find problems, or path is not a python module or script.
        """
        inspector = PathInspector(path)
        if inspector.is_toignore():
            return True
        if not inspector.is_python():
            return True

        try:
            runner = Run(self.get_opts() + [path], exit=False)
            if runner.linter.msg_status != 0:
                log.error('Pylint check fail: %s', path)
                self.failed_paths.append(path)
            return runner.linter.msg_status == 0
        except Exception, details:
            log.error('Pylint check fail: %s (pylint exception: %s)',
                      path, details)
            self.failed_paths.append(path)
            return False

    def check(self, path):
        if os.path.isfile(path):
            return self.check_file(path)
        elif os.path.isdir(path):
            return self.check_dir(path)


def set_arguments(parser):
    plint = parser.add_parser('lint', help='check code with pylint')
    plint.add_argument('path', type=str,
                       help='Path to check (empty for full tree check)',
                       nargs='*',
                       default=None)
    plint.set_defaults(func=run_lint)


def run_lint(args):
    paths = args.path
    if not paths:
        paths = [os.getcwd()]

    linter = Linter(verbose=args.verbose)

    status = True
    for path in paths:
        status &= linter.check(path)
    if status:
        log.info("Syntax check PASS")
        return 0
    else:
        log.error("Syntax check FAIL")
        return 1



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

import logging
import os
import sys

import pep8

try:
    import autopep8
    AUTOPEP8_CAPABLE = True
except ImportError:
    AUTOPEP8_CAPABLE = False

from inspector import PathInspector

log = logging.getLogger("inspektor.style")


class StyleChecker(object):

    def __init__(self, verbose=True):
        self.verbose = verbose
        self.ignored_errors = 'E501,E265,W601,E402'
        # Be able to analyze all imports inside the project
        sys.path.insert(0, os.getcwd())
        self.failed_paths = []

    def set_verbose(self):
        self.verbose = True

    def check_dir(self, path):
        """
        Recursively go on a directory checking files with PEP8.

        :param path: Path to a directory.
        """
        def visit(arg, dirname, filenames):
            for filename in filenames:
                self.check_file(os.path.join(dirname, filename))

        os.path.walk(path, visit, None)
        return not self.failed_paths

    def check_file(self, path):
        """
        Check one regular file with pylint for py syntax errors.

        :param path: Path to a regular file.
        :return: False, if pylint found syntax problems, True, if pylint didn't
                 find problems, or path is not a python module or script.
        """
        inspector = PathInspector(path)
        if inspector.is_toignore():
            return True
        if not inspector.is_python():
            return True
        try:
            opt_obj = pep8.StyleGuide().options
            ignore_list = self.ignored_errors.split(',') + list(opt_obj.ignore)
            opt_obj.ignore = tuple(set(ignore_list))
            # pylint: disable=E1123
            runner = pep8.Checker(filename=path, options=opt_obj)
        except:
            opts = ['--ignore'] + self.ignored_errors.split(',')
            pep8.process_options(opts)
            runner = pep8.Checker(filename=path)
        status = runner.check_all()
        if status != 0:
            log.error('PEP8 check fail: %s', path)
            self.failed_paths.append(path)
            if AUTOPEP8_CAPABLE:
                log.error('Trying to fix errors with autopep8')
                try:
                    opt_obj = autopep8.parse_args([path,
                                                   '--ignore',
                                                   self.ignored_errors,
                                                   '--in-place'])
                    autopep8.fix_file(path, opt_obj)
                except Exception, details:
                    log.error('Not able to fix errors: %s', details)
        return status == 0

    def check(self, path):
        if os.path.isfile(path):
            return self.check_file(path)
        elif os.path.isdir(path):
            return self.check_dir(path)


def set_arguments(parser):
    pstyle = parser.add_parser('style',
                               help='check code compliance to PEP8')
    pstyle.add_argument('path', type=str,
                        help='Path to check (empty for full tree check)',
                        nargs='*',
                        default=None)
    pstyle.set_defaults(func=run_style)


def run_style(args):
    paths = args.path
    if not paths:
        paths = [os.getcwd()]

    style_checker = StyleChecker(verbose=args.verbose)

    status = True
    for path in paths:
        status &= style_checker.check(path)
    if status:
        log.info("PEP8 compliance check PASS")
        return 0
    else:
        log.error("PEP8 compliance FAIL")
        return 1



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Base exception definitions.
"""


class CmdError(Exception):

    def __init__(self, command, result):
        self.command = command
        self.result = result

    def __str__(self):
        if self.result.exit_status is None:
            msg = "Command '%s' failed and is not responding to signals"
            msg %= self.command
        else:
            msg = "Command '%s' failed (rc=%d)"
            msg %= (self.command, self.result.exit_status)
        return msg



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Implements the base inspektor application.
"""
import logging
from argparse import ArgumentParser

from inspektor import lint
from inspektor import reindent
from inspektor import style
from inspektor import check
from inspektor import license

log = logging.getLogger("inspektor.app")


class InspektorApp(object):

    """
    Basic inspektor application.
    """

    def __init__(self):
        self.arg_parser = ArgumentParser(description='Inspektor code check')
        self.arg_parser.add_argument('-v', '--verbose', action='store_true',
                                     help='print extra debug messages',
                                     dest='verbose')

        subparsers = self.arg_parser.add_subparsers(title='subcommands',
                                                    description='valid subcommands',
                                                    help='subcommand help')
        lint.set_arguments(subparsers)
        reindent.set_arguments(subparsers)
        style.set_arguments(subparsers)
        check.set_arguments(subparsers)
        license.set_arguments(subparsers)
        self.args = self.arg_parser.parse_args()

    def run(self):
        return self.args.func(self.args)
